/* -*- pftp-c -*- */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <gtk/gtk.h>
#include <stdlib.h>
#include <stdio.h>
#include <inttypes.h>
#include <pftputil.h>
#include <pftp_client.h>
#include <stdlib.h>
#include <string.h>
#include "queue_view.h"
#include "list_view.h"
#include "globals.h"
#include "overwrite_dlg.h"

#ifdef WITH_DMALLOC
#include <dmalloc.h>
#endif

struct queue_view_s
{
    GtkTreeView *view;
    GtkListStore *store;
    pftp_que_t transfer_q, error_q;
};

static enum {
    ICON_COLUMN,
    SOURCE_COLUMN,
    TARGET_COLUMN,
    PROGRESS_COLUMN,
    N_COLUMNS
} queue_view_cols;

queue_view_t initQueueView(GtkTreeView *view)
{
    GtkCellRenderer *renderer, *iconrenderer;
    GtkTreeViewColumn *column;
    queue_view_t ret = malloc(sizeof(struct queue_view_s));
    memset(ret, 0, sizeof(struct queue_view_s));
    
    ret->view = view;
    ret->store = gtk_list_store_new(N_COLUMNS, G_TYPE_STRING, G_TYPE_STRING,
				    G_TYPE_STRING, G_TYPE_STRING);
    gtk_tree_view_set_model(ret->view, GTK_TREE_MODEL(ret->store)); 
    
    iconrenderer = gtk_cell_renderer_pixbuf_new();
    renderer = gtk_cell_renderer_text_new();
    column = gtk_tree_view_column_new_with_attributes("", iconrenderer, 
						      "stock-id", ICON_COLUMN, 
						      NULL);
    gtk_tree_view_column_set_min_width(column, 32);
    gtk_tree_view_append_column(ret->view, column);
    column = gtk_tree_view_column_new_with_attributes("Source", renderer, 
						      "text", SOURCE_COLUMN, 
						      NULL);
    gtk_tree_view_column_set_resizable(column, TRUE);
    gtk_tree_view_append_column(ret->view, column);
    column = gtk_tree_view_column_new_with_attributes("Target", renderer, 
						      "text",
						      TARGET_COLUMN, NULL);
    gtk_tree_view_column_set_resizable(column, TRUE);
    gtk_tree_view_append_column(ret->view, column);
    column = gtk_tree_view_column_new_with_attributes("Progress", renderer, 
						      "text", PROGRESS_COLUMN, 
						      NULL);
    gtk_tree_view_column_set_resizable(column, TRUE);
    gtk_tree_view_append_column(ret->view, column);
    
    ret->transfer_q = pftp_CreateQue();
    ret->error_q = pftp_CreateQue();
    
    return ret;
}

void freeQueueView(queue_view_t view)
{
    if (view) {
	pftp_FreeQue(view->transfer_q);
	pftp_FreeQue(view->error_q);
	free(view);   
    }
}

void updateQueueView(queue_view_t view);

void enqueueView(queue_view_t view, list_view_t src, list_view_t target, 
		 size_t count, list_view_file_t *item)
{
    size_t f = 0;
    int ret = 0;
    char *srcpath = NULL;
    pftp_directory_t **lsdata;
    pftp_server_t ftp = getListViewFTP(src, 1);
    lsdata = malloc(sizeof(pftp_directory_t *) * count);
    
    if (!ftp && getListViewName(src)) {
	output_printf(output, ERR, "Unable to login.\n");
	ret = -1;
    }

    if (pftp_get_path_dir(getListViewName(src), &srcpath))
	ret = -1;
    
    if (!ret) {
	for (f = 0; f < count; f++) {
	    if (!(lsdata[f] = pftp_trans_ls2(ftp, item[f].filename, 
					     NULL, 1, 0))) {
		output_printf(output, ERR, "Unable to list files.\n");
		ret = -1;
		break;
	    }
	}
    }
    
    if (!ret) {
	for (f = 0; f < count; f++) {
	    ret = pftp_get_files_ordered(getListViewName(src), 
					 getListViewName(target), srcpath,
					 lsdata[f], view->transfer_q, 1);
	    pftp_free_fdt((*lsdata[f]));
	    free(lsdata[f]);		    
	}
    } else {
	while (f > 0) {
	    pftp_free_fdt((*lsdata[f]));
	    free(lsdata[f]);
	}
    }

    if (srcpath)
	free(srcpath);
    free(lsdata);
    updateQueueView(view);
}

static void get_paths(pftp_item_t item, char **source, char **target);
static int safe_strcmp(const char *a, const char *b);

void updateQueueView(queue_view_t view)
{
    GtkTreeIter iter;
    pftp_item_t item;
    size_t pos = 0;
    char *source = NULL, *target = NULL;
    
    gtk_list_store_clear(view->store);
  
    while ((item = pftp_que_peek(view->error_q, pos++))) {
	get_paths(item, &source, &target);
	
	gtk_list_store_append(view->store, &iter);
	gtk_list_store_set(view->store, &iter, ICON_COLUMN, "gtk-error",
			   SOURCE_COLUMN, source, TARGET_COLUMN, target, 
			   PROGRESS_COLUMN, "<err>", -1);
    }
    
    pos = 0;
    
    while ((item = pftp_que_peek(view->transfer_q, pos++))) {
	get_paths(item, &source, &target);
	
	gtk_list_store_append(view->store, &iter);
	
	if (safe_strcmp(item->ftp_from, getListViewName(left)) == 0)
	    gtk_list_store_set(view->store, &iter, ICON_COLUMN, 
			       "gtk-go-forward",-1);
	else
	    gtk_list_store_set(view->store, &iter, ICON_COLUMN, 
			       "gtk-go-back", -1);
	
	gtk_list_store_set(view->store, &iter, SOURCE_COLUMN, source, 
			   TARGET_COLUMN, target, PROGRESS_COLUMN, "", -1);
    }
    
    gtk_tree_view_columns_autosize(view->view);
    
    if (source) free(source);
    if (target) free(target);
}

void get_paths(pftp_item_t item, char **source, char **target)
{
    size_t len1 = item->ftp_from ? strlen(item->ftp_from) + 1 : 0;
    size_t len2 = strlen(item->srcpath);
    size_t len3 = strlen(item->filename);
    
    (*source) = (char *) realloc((*source), len1 + len2 + 1 + len3 + 1);
    if (item->ftp_from) {
	memcpy((*source), item->ftp_from, len1 - 1);
	(*source)[len1 - 1] = ':';
    }
    memcpy((*source) + len1, item->srcpath, len2);
    (*source)[len1 + len2] = '/';
    memcpy((*source) + len1 + len2 + 1, item->filename, len3);
    (*source)[len1 + len2 + 1 + len3] = '\0';
    
    len1 = item->ftp_to ? strlen(item->ftp_to) + 1 : 0;
    len2 = strlen(item->destpath);
    len3 = item->target ? strlen(item->target) : len3;
    
    (*target) = (char *) realloc((*target), len1 + len2 + 1 + len3 + 1);
    if (item->ftp_to) {
	memcpy((*target), item->ftp_to, len1 - 1);
	(*target)[len1 - 1] = ':';
    }
    memcpy((*target) + len1, item->destpath, len2);
    (*target)[len1 + len2] = '/';
    if (item->target) {
	memcpy((*target) + len1 + len2 + 1, item->target, len3);
	(*target)[len1 + len2 + 1 + len3] = '\0';
    } else {
	memcpy((*target) + len1 + len2 + 1, item->filename, len3);
	(*target)[len1 + len2 + 1 + len3] = '\0';
    }
}

int safe_strcmp(const char *a, const char *b)
{
    if (a) {
	if (b) {
	    return strcmp(a, b);
	} else {
	    return -1;
	}
    } else {
	if (b) {
	    return 1;
	} else {
	    return 0;
	}
    }
}

void startQueueView(queue_view_t view)
{
    static int que_running = 0;
    if (que_running)
	return;
    abort_transfer = 0;
    abort_que = 0;
    que_running = 1;
    pftp_begin_trans(NULL, overwrite_dlg, view->transfer_q, view->error_q);
    que_running = 0;
    updateListViewDir(left);
    updateListViewDir(right);
}

void abortQueueView(queue_view_t view)
{
    if (!que_running)
	return;

    abort_transfer = 1;
    abort_que = 1;
}

void pauseQueueView(queue_view_t view)
{
    if (!que_running)
	return;

    abort_que = 1;
}

pftp_server_t getQueFtpTarget(queue_view_t view)
{
    pftp_item_t q = pftp_que_peek(view->transfer_q, 0);
    if (q && q->ftp_to) {
	if (getListViewName(left) 
	    && strcmp(q->ftp_to, getListViewName(left)) == 0)
	    return getListViewFTP(left, 0);
	if (getListViewName(right) && 
	    strcmp(q->ftp_to, getListViewName(right)) == 0)
	    return getListViewFTP(right, 0);
    }
    
    return NULL;
}

void dequeueView(queue_view_t view)
{
    /* !!! */
}
